// ComClass.cpp this is the implementation off the graphicstablet

#include "comclass.h"
#include "tty.h"
#include <ctype.h>
#include <time.h>
#include <strclass.h>
#include "interror.h"

               
void ComClass::PutChar(char c)
{
    int r;    
    time_t t_a, t_b;            
    double dur;
                     
    if(!ok) return; 
    
    r=WriteComm(idComDev, &c, 1);
    t_a=time(NULL);
    
    do {
      if(r!=1 || (errno=GetCommError(idComDev, &cmst))!=0) { ok=0; }
      t_b=time(NULL);
      dur=difftime(t_b,t_a);
    } while (ok && cmst.cbOutQue>1 && dur<TimeOut_secs);  
    if((TimeOut_Flag = dur >= TimeOut_secs)!=0) { ok=0; }
}

char ComClass::GetChar()
{
    int r;     
    time_t t_a, t_b;            
    double dur;
    char c;
    
    if(!ok) return 0; 
        
    t_a=time(NULL);
    do
    {
      if((errno=GetCommError(idComDev, &cmst))!=0) { ok=0; }
      t_b=time(NULL);
      dur=difftime(t_b,t_a);
    } while (ok && cmst.cbInQue<1 && dur<TimeOut_secs);  
    if((TimeOut_Flag = dur >= TimeOut_secs)!=0) { ok=0; }
    if(ok)
    {
      r=ReadComm(idComDev, &c, 1);
      ok= r==1 ;
    }
    return c;
}
                             
void ComClass::OpenPort(int port_num, int baud, char parity, 
												int bits, bool hardflow, bool binary,
												double stopbits
                       )
{
  BYTE bset ;
  DCB dcb;
  char Buffer[100];
  char Buffer2[10];
  int r, key;
  size_t i, j;

  sprintf(Buffer,"COM%c", '0'+port_num);
  
  idComDev = OpenComm(Buffer, 1024, 128);
  
  if (idComDev < 0) 
  {
      ok=0;
      return;
  }
                   
  if(((int)(stopbits*10))%10)
    sprintf(Buffer2,".%d", ((int)(stopbits*10))%10);
  else
    Buffer2[0]=0;
      
  sprintf(Buffer,"COM%c:%d,%c,%d,%d%s", '0'+port_num, baud, parity, 
          bits, (int)stopbits, Buffer2);
  
  int err = BuildCommDCB(Buffer, &dcb);
  if (err < 0) 
  {
      ok=0;
      return;
  }
  
  if(parity=='O')
    dcb.Parity   = ODDPARITY;                               
  else if(parity=='E')
    dcb.Parity   = EVENPARITY;                               
  else if(parity=='N')
    dcb.Parity   = NOPARITY;                               
  else if(parity=='M')
    dcb.Parity   = MARKPARITY;                               
  else if(parity=='S')
    dcb.Parity   = SPACEPARITY;
  else INT_ERR(__LINE__);                                

  switch(baud)
  {
    case 110: dcb.BaudRate = CBR_110;
              break;
    case 300: dcb.BaudRate = CBR_300;
              break;
    case 600: dcb.BaudRate = CBR_600;
              break;
    case 1200: dcb.BaudRate = CBR_1200;
              break;
    case 2400: dcb.BaudRate = CBR_2400;
              break;
    case 4800: dcb.BaudRate = CBR_4800;
              break;
    case 9600: dcb.BaudRate = CBR_9600;
              break;
    case 14400: dcb.BaudRate = CBR_14400;
              break;
    case 19200: dcb.BaudRate = CBR_19200;
              break;
    case 38400: dcb.BaudRate = CBR_38400;
              break;
    case 56000: dcb.BaudRate = CBR_56000;
              break;
    case 128000: dcb.BaudRate = CBR_128000;
              break;
    case 256000: dcb.BaudRate = CBR_256000;
              break;                      
    default : INT_ERR(__LINE__);
  }            
  dcb.ByteSize = bits;
  if(stopbits==1.0)
    dcb.StopBits = ONESTOPBIT;
  else if(stopbits==1.5)
    dcb.StopBits = ONE5STOPBITS;
  else if(stopbits==2.0)
    dcb.StopBits = TWOSTOPBITS;
  else INT_ERR(__LINE__);  
                            

  // setup hardware flow control

  //bSet = (BYTE) ((FLOWCTRL( npTTYInfo ) & FC_DTRDSR) != 0) ;
	bset = hardflow;
  dcb.fOutxDsrFlow = dcb.fDtrflow = bset ;
  dcb.DsrTimeout = (bset) ? 30 : 0 ;

  // bSet = (BYTE) ((FLOWCTRL( npTTYInfo ) & FC_RTSCTS) != 0) ;
	bset = hardflow;
  dcb.fOutxCtsFlow = dcb.fRtsflow = bset ;
  dcb.CtsTimeout = (bset) ? 30 : 0 ;

  // setup software flow control

  // bSet = (BYTE) ((FLOWCTRL( npTTYInfo ) & FC_XONXOFF) != 0) ;
	bset = hardflow==0;
  dcb.fInX = dcb.fOutX = bset;
  dcb.XonChar = ASCII_XON ;
  dcb.XoffChar = ASCII_XOFF ;
  dcb.XonLim = 100 ;
  dcb.XoffLim = 100 ;

  // other various settings

	dcb.fBinary = binary ;
  dcb.fParity = (parity=='N')?0:1 ;
	dcb.fRtsDisable = (hardflow==0)?1:0 ;
	dcb.fDtrDisable = (hardflow==0)?1:0 ;

  
  err = SetCommState(&dcb);
  if (err < 0) 
  {
      ok=0;
      return;
  }
                        
  // Flush the receiving queue, because of trash-data                                         
  ok=FlushComm(idComDev, 1)>=0;
  if(!ok || (errno=GetCommError(idComDev, &cmst))!=0) { ok=0; return; }
  
  // Flush the transmitter queue, because of trash-data                                         
  ok=FlushComm(idComDev, 0)>=0;
  if(!ok || (errno=GetCommError(idComDev, &cmst))!=0) { ok=0; return; }
  
  ok=1;
}

void ComClass::ClosePort()
{
  CloseComm(idComDev);
}

