// author : Sven Goethel - Stapenhorststrasse 35a - 33615 Bielefeld
//
/*************************************************************************/
/*
 *  Datei          : metageom.h
 *  Programmierer  : Sven Goethel
 *            Stapenhorststr.35a
 *            33615 Bielefeld
 *            0521/139228
 *
 * $Log: metageom.h $
 * Revision 1.2  1995/04/07 16:54:54  Sven_Goethel_DFP
 * *** empty log message ***
 * Revision 1.1  1995/04/04 10:37:31  Sven_Goethel
 * Initial revision
 */
/*************************************************************************/

# ifndef _METAGEOM_H
  # define _METAGEOM_H

  # include "metatype.h"
  # include <math.h>

  class MetaGeometry 
  {
    public:
      enum Position { Ende=0, Anfang=1 };
    static t_real AddWinkel(const t_real a, const t_real b);
    static t_real SubWinkel(const t_real a, const t_real b);

    static t_real VektorBetrag ( const t_coord &P0, const t_coord &PM,
                                 const t_coord &P1, t_rotation rot);
    static t_real VektorBetrag ( const t_coord &P1, const t_coord &P2);

    static t_real GetWinkelAbsolut ( const t_coord &Pp, const t_coord &PM,
                                     t_rotation Rot) ;
    static t_real GetWinkelAbsolut ( const t_coord &, const t_coord & );

    // cosinus, eingeschl Winkel zwisch. P0->P1, P1->P2
    //  OK :: 0<= return <= 360 
    // !OK :: return < 0 
    static t_real GetWinkel12 ( const t_coord &P0, const t_coord &P1,
                                const t_coord &P2
                              );  
    static t_real GetWinkel12 ( const t_coord &P0, const t_coord &P1,
                                const t_coord &P2, t_rotation rot
                              );
    static t_real GetWinkel12 ( t_real Alpha, t_real Beta,
                                t_rotation rot
                              );    
                              
    static t_rotation CalcRotation( const t_coord & p1, const t_coord & p2, 
                                    const t_coord & p3 
                                  );
    // Umwandlung der Kreis-Formate
    //
    //    PPP := Drei unterschiedliche Punkte auf einem Kreis,
    //    PMP := StartPunkt, MittelPunkt, EndPunkt, Drehrichtung (CW || CCW)
    //
    // Falls bei der Eingabe keine unterschiedichen Punkte
    // uebergeben werden => return false !!!!!                          
    // 
    // Der PPP Rueckgabewert verhaelt sich wie folgt :
    //    Winkel ( P1->P2 , P2->P3 ) == 2 * Winkel ( P1->P2 , P2->PM )
    //    wobei das '->' bedeutet Strecke von Links-Punkt zu Rechts-Punkt !!!
    //
    static bool ArcConvertPPP2PMP ( const t_coord & p1, const t_coord & p2, 
                                    const t_coord & p3,             
                                    t_coord & pS, t_coord & pM, 
                                    t_coord & pE, t_rotation &rot
                                  );
                                  
    static bool ArcConvertPMP2PPP ( const t_coord & pS, const t_coord & pM, 
                                    const t_coord & pE, t_rotation rot,             
                                    t_coord & p1, t_coord & p2, 
                                    t_coord & p3
                                  );               
                                  
    static bool TestBogenPunkt (const t_coord &ap, const t_coord &mp, 
                                const t_coord &ep, t_rotation rot, 
                                const t_coord &p
                               ) ;
                               
    // Point: P, Line: A + s * B ( s = any scalar )                                         
    static t_real DistancePointLine(const t_coord &P, 
                                    const t_coord &A, const t_coord &B
                                   );
                               
    static bool TestStreckenPunkt (const t_coord &pS, const t_coord &pE, 
                                   const t_coord &p
                                  );
    // StreckeA := PA0 -> PA1              
    // StreckeB := PB0 -> PB1              
    // return   := Ein Schnittpunkt existiert, nicht unbedingt auf den Strecken !!!!
    static bool SchnittStreckeStrecke(const t_coord &PA0, const t_coord &PA1, 
                                      const t_coord &PB0, const t_coord &PB1,
                                      t_coord &SchnittPunkt, 
                                      bool &AufStreckeA,
                                      bool &AufStreckeB
                                     );
  
    // Gerade a := Oa + Ga * s                            
    // Gerade b := Ob + Gb * t                            
    // return   := Ein Schnittpunkt existiert !!!
    static bool SchnittGrdGrd (const t_coord &Oa, const t_coord &Ga, 
                               const t_coord &Ob, const t_coord &Gb,
                               t_coord &SchnittPunkt
                              );
                              

    // Strecke := pA0 -> pA1              
    // Bogen   := pBS , pBM , pBE , rot ; mit dem radius := (pBS-pBM).length() !!              
    // return   := Anzahl d. Schnittpunkte!!!!
    // Be sure that the SchnittPunkte-Array, the AufStrecke-Array, 
    // and the the AufBogen-Array, is of the size 2 !!!!! 
    static int  SchnittStreckeBogen  (const t_coord &pA0, const t_coord &pA1, 
                                      const t_coord &pBS, const t_coord &pBM,
                                      const t_coord &pBE, t_rotation rot,
                                      t_coord SchnittPunkte[], 
                                      bool AufStrecke[], bool AufBogen[]
                                     );
  
    // O = [ o1,o2,o3 ] ; Der Ortsvektor der Geradengleichung
    // G = [ g1,g2,g3 ] ; Der Richtungsvektor der Geradengleichung
    // M = [ m1,m2,m3 ] ; Der Kreismittelpunkt 
    // r                ; Der Radius
    // return           ; Anzahl der Schnittpunkte !!!
    // Be sure that the SchnittPunkte-Array is of the size 2 !!!!! 
    static int SchnittGrdKreis (const t_coord &O, const t_coord &G, 
                                const t_coord &M, const t_real &radius,
                                t_coord SchnittPunkte[]
                               );

    // Bogen1   := pB1S , pB1M , pB1E , B1rot ; radius := (pB1S-pB1M).length() !!              
    // Bogen2   := pB2S , pB2M , pB2E , B2rot ; radius := (pB2S-pB2M).length() !!              
    // return   := Anzahl d. Schnittpunkte!!!!
    // Be sure that the SchnittPunkte-Array, the AufStrecke-Array, 
    // and the the AufBogen-Array, is of the size 2 !!!!! 
    static int SchnittBogenBogen(
                                  const t_coord &pB1S, const t_coord &pB1M,
                                  const t_coord &pB1E, t_rotation B1rot,
                                  const t_coord &pB2S, const t_coord &pB2M,
                                  const t_coord &pB2E, t_rotation B2rot,
                                  t_coord SchnittPunkte[], 
                                  bool AufBogen1[], bool AufBogen2[]
                                );
    // M1=[ m1,m2,m3 ] ; Der 1. Kreismittelpunkt ( m3 bleibt unberuecksichtigt )
    // M1=[ m1,m2,m3 ] ; Der 2. Kreismittelpunkt ( m3 bleibt unberuecksichtigt )
    // r1, r2           ; Die beiden Radien
    // return           ; Anzahl der Schnittpunkte !!!             
    //                  ; ist return == 0 -> Keine SchnittPunkte
    //                  ; ist return <= 2 -> 1, bzw 2 SchnittPunkte 
    //                  ;   Ergebniss in SchnittPunkte[] 
    //                  ; ist return == 3 -> Identische Kreise 
    //                  ;   SchnittPunkte[] unveraendert !!
    // Be sure that the SchnittPunkte-Array is of the size 2 !!!!! 
    static int SchnittKreisKreis (const t_coord &M1, const t_real &r1, 
                                  const t_coord &M2, const t_real &r2, 
                                  t_coord SchnittPunkte[]
                                 );
  };

#endif


